@@ -3,6 +3,7 @@ from __future__ import annotations
 import dataclasses
 
 from typing import Optional
+from typing import Sequence
 from typing import Tuple
 from typing import Union
 
@@ -45,7 +46,7 @@ class Release:
                 object.__setattr__(self, "patch", 0)
         parts = [
             str(part)
-            for part in [self.major, self.minor, self.patch, *self.extra]
+            for part in (self.major, self.minor, self.patch, *self.extra)
             if part is not None
         ]
         object.__setattr__(self, "text", ".".join(parts))
@@ -68,6 +69,13 @@ class Release:
             extra=parts[3:],
         )
 
+    def to_parts(self) -> Sequence[int]:
+        return tuple(
+            part
+            for part in [self.major, self.minor, self.patch, *self.extra]
+            if part is not None
+        )
+
     def to_string(self) -> str:
         return self.text
 
@@ -98,6 +106,24 @@ class Release:
             extra=tuple(0 for _ in self.extra),
         )
 
+    def next(self) -> Release:
+        if self.precision == 1:
+            return self.next_major()
+
+        if self.precision == 2:
+            return self.next_minor()
+
+        if self.precision == 3:
+            return self.next_patch()
+
+        return dataclasses.replace(
+            self,
+            major=self.major,
+            minor=self.minor,
+            patch=self.patch,
+            extra=(*self.extra[:-1], self.extra[-1] + 1),
+        )
+
 
 @dataclasses.dataclass(frozen=True, eq=True, order=True)
 class ReleaseTag:
